from __future__ import print_function

from gomjabbar.const import (
    NAME_TABLE, PARAMETER_TABLE, PORT_TABLE, STATIC_VAR_TABLE,
    ALLOWED_TYPES, ARRAY, ARRAY_BOUNDS, DATA_TYPE, DEFAULT_TYPE, DEFAULT_VALUE,
    DESCRIPTION, DIRECTION, LIMITS, NULL_ALLOWED, PARAMETER_NAME, PORT_NAME,
    STATIC_VAR_NAME
)
from .ast import Dash, Ifs, Table, TableRow
from .lexer import IfsLexer
from .parser import IfsParser

PORT_TYPES = {
    'v': 'MIF_VOLTAGE',
    'vd': 'MIF_DIFF_VOLTAGE',
    'i': 'MIF_CURRENT',
    'id': 'MIF_DIFF_CURRENT',
    'vnam': 'MIF_VSOURCE_CURRENT',
    'g': 'MIF_CONDUCTANCE',
    'gd': 'MIF_DIFF_CONDUCTANCE',
    'h': 'MIF_RESISTANCE',
    'hd': 'MIF_DIFF_RESISTANCE',
    'd': 'MIF_DIGITAL',
}
VALUE_UNION_NAMES = {
    'boolean': (bool, 'bvalue'),
    'int': (int, 'ivalue'),
    'real': (float, 'rvalue'),
    'complex': (lambda x: x, 'cvalue'),
    'string': (lambda x: '"{}"'.format(x), 'svalue'),
    'pointer': (lambda x: x, 'pvalue'),
}


class _Namespace(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


def build_connections_list(port_table):
    """ Convert the PORT_TABLE items into a list which can supply data to the
    connections code template.
    """
    connections = []
    for item in transpose_table(port_table):
        conn = {
            'name': item.PORT_NAME.value,
            'is_input': item.DIRECTION.value in ('in', 'inout'),
            'is_output': item.DIRECTION.value in ('out', 'inout'),
            'ports': [
                {'type': PORT_TYPES.get(item.DEFAULT_TYPE.value,
                                        'MIF_USER_DEFINED')}
            ],
        }
        connections.append(conn)
    return connections


def build_parameters_list(parameter_table, parameter_values):
    """ Convert the PARAMETER_TABLE items into a list which can supply data to
    the params code template.
    """
    parameters = []
    for item in transpose_table(parameter_table):
        param_name = item.PARAMETER_NAME.value
        cast, unionmember = VALUE_UNION_NAMES[item.DATA_TYPE.value]
        param = {
            'name': param_name,
            'unionmember': unionmember,
        }
        value = parameter_values.get(param_name, item.DEFAULT_VALUE)
        if not isinstance(value, Dash):
            if isinstance(value, list):
                value = [cast(v) for v in value]
            else:
                value = [cast(value)]
            param['values'] = value
        parameters.append(param)
    return parameters


def compact_ast(ifs_ast):
    """ Group all tables in an `Ifs` object by their type.

    NOTE: If there is more than one NAME_TABLE, only the final one given in
    the .ifs source file will be kept.
    """
    tables = {}
    compacted = []
    ifs_kwargs = {}

    for tab in ifs_ast:
        tables.setdefault(tab.type, []).append(tab)

    # The last NAME_TABLE is the one which is used
    name_tabs = tables.get(NAME_TABLE, [])
    if name_tabs:
        compacted.append(name_tabs[-1])
        ifs_kwargs['name_table'] = name_tabs[-1]

    compaction_data = {
        PARAMETER_TABLE: {ARRAY, ARRAY_BOUNDS, DATA_TYPE, DEFAULT_VALUE,
                          DESCRIPTION, LIMITS, NULL_ALLOWED, PARAMETER_NAME},
        PORT_TABLE: {ALLOWED_TYPES, ARRAY, ARRAY_BOUNDS, DEFAULT_TYPE,
                     DESCRIPTION, DIRECTION, NULL_ALLOWED, PORT_NAME},
        STATIC_VAR_TABLE: {ARRAY, DATA_TYPE, DESCRIPTION, STATIC_VAR_NAME},
    }
    for name, expected_rows in compaction_data.items():
        if name in tables:
            table = compact_tables(name, tables[name], expected_rows)
            compacted.append(table)
            ifs_kwargs[name.lower()] = table

    return Ifs(compacted, **ifs_kwargs)


def compact_tables(table_type, tables, expected_row_names):
    """ Collapse a list of tables of the same type into a single table
    """
    rows = {}
    for tab in tables:
        expected_rows = expected_row_names.copy()
        col_count = len(tab.nodes[0].value)
        for row in tab:
            rows.setdefault(row.type, []).extend(row.value)
            expected_rows.remove(row.type)

        for typename in expected_rows:
            value = [None] * col_count
            rows.setdefault(row.type, []).extend(value)

    return Table(table_type, [TableRow(k, v) for k, v in rows.items()])


def count_static_vars(static_vars_table):
    """ Count the number of variables defined in the STATIC_VARS_TABLE
    """
    return len(static_vars_table.nodes[0].value)


def parse_file(path):
    """ Parse a single .ifs file and return its AST
    """
    parser = IfsParser()
    with open(path, 'r') as fp:
        ast = parser.parse(fp.read())
    return compact_ast(ast)


def tokenize_file(path):
    """ Show the result of running the IfsLexer on a file.
    """
    lexer = IfsLexer()
    with open(path, 'r') as fp:
        lexer.lexer.input(fp.read())

    while True:
        tok = lexer.lexer.token()
        if tok is None:
            break
        print(tok)


def transpose_table(table):
    """ Convert a `Table` to a list of objects which represent the "columns".
    """
    items = [{} for _ in range(len(table.nodes[0].value))]
    for row in table:
        for i, value in enumerate(row.value):
            items[i][row.type] = value
    return [_Namespace(**d) for d in items]


if __name__ == '__main__':
    import argparse

    argparser = argparse.ArgumentParser()
    argparser.add_argument('-l', '--lex', type=str, default='')
    argparser.add_argument('-p', '--parse', type=str, default='')
    args = argparser.parse_args()

    if args.lex:
        tokenize_file(args.lex)
    elif args.parse:
        print(parse_file(args.parse))




class AstNode(object):
    def get_children(self):
        return ()

    def __iter__(self):
        for c in self.get_children():
            yield c


class CType(AstNode):
    def __init__(self, value, lineno=None):
        self.value = value
        self.lineno = lineno

    def __repr__(self):
        return 'CType({})'.format(self.value)


class Dash(AstNode):
    def __init__(self, lineno=None):
        self.lineno = lineno

    def __repr__(self):
        return 'Dash'


class Direction(AstNode):
    def __init__(self, value, lineno=None):
        self.value = value
        self.lineno = lineno

    def __repr__(self):
        return 'Direction({})'.format(self.value)


class DType(AstNode):
    def __init__(self, value, lineno=None):
        self.value = value
        self.lineno = lineno

    def __repr__(self):
        return 'DType({})'.format(self.value)


class Identifier(AstNode):
    def __init__(self, value, lineno=None):
        self.value = value
        self.lineno = lineno

    def __repr__(self):
        return 'Identifier({})'.format(self.value)


class Ifs(AstNode):
    def __init__(self, nodes, name_table=None, parameter_table=None,
                 port_table=None, static_var_table=None):
        self.nodes = nodes
        self.name_table = name_table
        self.parameter_table = parameter_table
        self.port_table = port_table
        self.static_var_table = static_var_table
        self.lineno = 1

    def get_children(self):
        return self.nodes

    def __repr__(self):
        nodes = '\n'.join('{!r}'.format(n) for n in self)
        if nodes:
            nodes = '\n' + nodes + '\n'
        return 'Ifs({})'.format(nodes)


class Range(AstNode):
    def __init__(self, low=None, high=None, lineno=None):
        self.low = Dash(lineno=lineno) if low is None else low
        self.high = Dash(lineno=lineno) if high is None else high
        self.lineno = lineno

    def __repr__(self):
        return 'Range({!r}, {!r})'.format(self.low, self.high)


class Table(AstNode):
    def __init__(self, type_, nodes, lineno=None):
        self.type = type_
        self.nodes = nodes
        self.lineno = lineno

    def get_children(self):
        return self.nodes

    def __repr__(self):
        nodes = '\n'.join('\t{!r}'.format(n) for n in self)
        if nodes:
            nodes = '\n' + nodes + '\n'
        return '{}({})'.format(self.type, nodes)


class TableRow(AstNode):
    def __init__(self, type_, value, lineno=None):
        self.type = type_
        self.value = value
        self.lineno = lineno

    def __repr__(self):
        return '{}({!r})'.format(self.type, self.value)




NAME_TABLE = 'NAME_TABLE'
PARAMETER_TABLE = 'PARAMETER_TABLE'
PORT_TABLE = 'PORT_TABLE'
STATIC_VAR_TABLE = 'STATIC_VAR_TABLE'

ALLOWED_TYPES = 'ALLOWED_TYPES'
ARRAY = 'ARRAY'
ARRAY_BOUNDS = 'ARRAY_BOUNDS'
C_FUNCTION_NAME = 'C_FUNCTION_NAME'
DATA_TYPE = 'DATA_TYPE'
DEFAULT_TYPE = 'DEFAULT_TYPE'
DEFAULT_VALUE = 'DEFAULT_VALUE'
DESCRIPTION = 'DESCRIPTION'
DIRECTION = 'DIRECTION'
LIMITS = 'LIMITS'
NULL_ALLOWED = 'NULL_ALLOWED'
PARAMETER_NAME = 'PARAMETER_NAME'
PORT_NAME = 'PORT_NAME'
SPICE_MODEL_NAME = 'SPICE_MODEL_NAME'
STATIC_VAR_NAME = 'STATIC_VAR_NAME'



from __future__ import print_function

import os.path as op
import re

import ply.lex as lex

from .ast import CType, Direction, DType, Identifier


class IfsLexer(object):
    """ Lexer for .ifs files
    """
    states = (
        ('bool', 'inclusive'),
        ('ctype', 'inclusive'),
        ('dir', 'inclusive'),
        ('dtype', 'inclusive'),
        ('comment', 'exclusive'),
        ('stringl', 'exclusive'),
    )
    tokens = (
        'STRING_LITERAL',
        'ALLOWED_TYPES',
        'ARRAY',
        'ARRAY_BOUNDS',
        'C_FUNCTION_NAME',
        'PORT_NAME',
        'PORT_TABLE',
        'DATA_TYPE',
        'DEFAULT_TYPE',
        'DEFAULT_VALUE',
        'DESCRIPTION',
        'DIRECTION',
        'STATIC_VAR_NAME',
        'STATIC_VAR_TABLE',
        'LIMITS',
        'NAME_TABLE',
        'NULL_ALLOWED',
        'PARAMETER_NAME',
        'PARAMETER_TABLE',
        'SPICE_MODEL_NAME',

        'BOOL_YES',
        'BOOL_NO',

        'CTYPE_V',
        'CTYPE_VD',
        'CTYPE_VNAM',
        'CTYPE_I',
        'CTYPE_ID',
        'CTYPE_G',
        'CTYPE_GD',
        'CTYPE_H',
        'CTYPE_HD',
        'CTYPE_D',

        'DIR_IN',
        'DIR_OUT',
        'DIR_INOUT',

        'DTYPE_REAL',
        'DTYPE_INT',
        'DTYPE_BOOLEAN',
        'DTYPE_COMPLEX',
        'DTYPE_STRING',
        'DTYPE_POINTER',

        'LANGLE',
        'RANGLE',
        'LBRACKET',
        'RBRACKET',
        'COMMA',
        'DASH',

        'IDENTIFIER',
        'INT_LITERAL',
        'REAL_LITERAL',
    )

    def t_start_comment_state(self, t):
        r'/\*'
        t.lexer.push_state('comment')

    def t_comment_non_star(self, t):
        r'[^*\n]+'
        # eat anything that's not a '*'
        pass

    def t_comment_stars(self, t):
        r'\*(?!/)'
        # eat up '*'s not followed by '/'s
        pass

    def t_comment_nl(self, t):
        r'\n'
        # new line
        t.lexer.lineno += 1

    def t_comment_end(self, t):
        r'\*+/'
        t.lexer.pop_state()

    # supress PLY warning
    t_comment_ignore = ''

    def t_comment_error(self, t):
        raise RuntimeError('Error while scanning comment')

    def t_start_stringl_state(self, t):
        r'"'
        t.lexer.push_state('stringl')

    def t_stringl_content(self, t):
        r'[^"]+'
        t.type = 'STRING_LITERAL'
        return t

    def t_stringl_end(self, t):
        r'"'
        t.lexer.pop_state()

    # supress PLY warning
    t_stringl_ignore = ''

    def t_stringl_error(self, t):
        raise RuntimeError('Error while scanning string literal')

    def t_ARRAY_BOUNDS(self, t):
        r'vector_bounds[ \t\n]*:'
        t.value = t.type
        return t

    def t_C_FUNCTION_NAME(self, t):
        r'c_function_name[ \t\n]*:'
        t.value = t.type
        return t

    def t_PORT_NAME(self, t):
        r'port_name[ \t\n]*:'
        t.value = t.type
        return t

    def t_PORT_TABLE(self, t):
        r'port_table[ \t\n]*:'
        t.value = t.type
        return t

    def t_DEFAULT_VALUE(self, t):
        r'default_value[ \t\n]*:'
        t.value = t.type
        return t

    def t_DESCRIPTION(self, t):
        r'description[ \t\n]*:'
        t.value = t.type
        return t

    def t_STATIC_VAR_NAME(self, t):
        r'static_var_name[ \t\n]*:'
        t.value = t.type
        return t

    def t_STATIC_VAR_TABLE(self, t):
        r'static_var_table[ \t\n]*:'
        t.value = t.type
        return t

    def t_LIMITS(self, t):
        r'limits[ \t\n]*:'
        t.value = t.type
        return t

    def t_NAME_TABLE(self, t):
        r'name_table[ \t\n]*:'
        t.value = t.type
        return t

    def t_PARAMETER_NAME(self, t):
        r'parameter_name[ \t\n]*:'
        t.value = t.type
        return t

    def t_PARAMETER_TABLE(self, t):
        r'parameter_table[ \t\n]*:'
        t.value = t.type
        return t

    def t_SPICE_MODEL_NAME(self, t):
        r'spice_model_name[ \t\n]*:'
        t.value = t.type
        return t

    def t_ALLOWED_TYPES(self, t):
        r'allowed_types[ \t\n]*:'
        t.value = t.type
        t.lexer.push_state('ctype')
        return t

    def t_ARRAY(self, t):
        r'vector[ \t\n]*:'
        t.value = t.type
        t.lexer.push_state('bool')
        return t

    def t_DATA_TYPE(self, t):
        r'data_type[ \t\n]*:'
        t.value = t.type
        t.lexer.push_state('dtype')
        return t

    def t_DEFAULT_TYPE(self, t):
        r'default_type[ \t\n]*:'
        t.value = t.type
        t.lexer.push_state('ctype')
        return t

    def t_DIRECTION(self, t):
        r'direction[ \t\n]*:'
        t.value = t.type
        t.lexer.push_state('dir')
        return t

    def t_NULL_ALLOWED(self, t):
        r'null_allowed[ \t\n]*:'
        t.value = t.type
        t.lexer.push_state('bool')
        return t

    def t_BOOL_YES(self, t):
        r'true'
        t.value = True
        return t

    def t_BOOL_NO(self, t):
        r'false'
        t.value = False
        return t

    def t_bool_BOOL_YES(self, t):
        r'yes'
        t.value = True
        return t

    def t_bool_BOOL_NO(self, t):
        r'no'
        t.value = False
        return t

    def t_bool_end(self, t):
        r'\n'
        t.lexer.lineno += 1
        t.lexer.pop_state()

    def t_ctype_CTYPE_V(self, t):
        r'v(?!d|nam)'
        t.value = CType(t.value, lineno=t.lexer.lineno)
        return t

    def t_ctype_CTYPE_VD(self, t):
        r'vd'
        t.value = CType(t.value, lineno=t.lexer.lineno)
        return t

    def t_ctype_CTYPE_VNAM(self, t):
        r'vnam'
        t.value = CType(t.value, lineno=t.lexer.lineno)
        return t

    def t_ctype_CTYPE_I(self, t):
        r'i(?!d)'
        t.value = CType(t.value, lineno=t.lexer.lineno)
        return t

    def t_ctype_CTYPE_ID(self, t):
        r'id'
        t.value = CType(t.value, lineno=t.lexer.lineno)
        return t

    def t_ctype_CTYPE_G(self, t):
        r'g(?!d)'
        t.value = CType(t.value, lineno=t.lexer.lineno)
        return t

    def t_ctype_CTYPE_GD(self, t):
        r'gd'
        t.value = CType(t.value, lineno=t.lexer.lineno)
        return t

    def t_ctype_CTYPE_H(self, t):
        r'h(?!d)'
        t.value = CType(t.value, lineno=t.lexer.lineno)
        return t

    def t_ctype_CTYPE_HD(self, t):
        r'hd'
        t.value = CType(t.value, lineno=t.lexer.lineno)
        return t

    def t_ctype_CTYPE_D(self, t):
        r'd'
        t.value = CType(t.value, lineno=t.lexer.lineno)
        return t

    def t_ctype_end(self, t):
        r'\n'
        t.lexer.lineno += 1
        t.lexer.pop_state()

    def t_dir_DIR_IN(self, t):
        r'in(?!out)'
        t.value = Direction(t.value)
        return t

    def t_dir_DIR_OUT(self, t):
        r'out'
        t.value = Direction(t.value)
        return t

    def t_dir_DIR_INOUT(self, t):
        r'inout'
        t.value = Direction(t.value)
        return t

    def t_dir_end(self, t):
        r'\n'
        t.lexer.lineno += 1
        t.lexer.pop_state()

    def t_dtype_DTYPE_REAL(self, t):
        r'real'
        t.value = DType(t.value, lineno=t.lexer.lineno)
        return t

    def t_dtype_DTYPE_INT(self, t):
        r'int'
        t.value = DType(t.value, lineno=t.lexer.lineno)
        return t

    def t_dtype_DTYPE_BOOLEAN(self, t):
        r'boolean'
        t.value = DType(t.value, lineno=t.lexer.lineno)
        return t

    def t_dtype_DTYPE_COMPLEX(self, t):
        r'complex'
        t.value = DType(t.value, lineno=t.lexer.lineno)
        return t

    def t_dtype_DTYPE_STRING(self, t):
        r'string'
        t.value = DType(t.value, lineno=t.lexer.lineno)
        return t

    def t_dtype_DTYPE_POINTER(self, t):
        r'pointer'
        t.value = DType(t.value, lineno=t.lexer.lineno)
        return t

    def t_dtype_end(self, t):
        r'\n'
        t.lexer.lineno += 1
        t.lexer.pop_state()

    t_LANGLE = r'<'
    t_RANGLE = r'>'
    t_LBRACKET = r'\['
    t_RBRACKET = r'\]'
    t_COMMA = r','
    t_DASH = r'-'

    def t_IDENTIFIER(self, t):
        r'[a-z_]+\w*'
        t.value = Identifier(t.value, lineno=t.lexer.lineno)
        return t

    def t_REAL_LITERAL(self, t):
        r'''[+-]?\d+\.\d*(e[+-]?\d+)?|
            [+-]?\d*\.\d+(e[+-]?\d+)?|
            [+-]?\d+(e[+-]?\d+)'''
        t.value = float(t.value)
        return t

    def t_INT_LITERAL(self, t):
        r'[+-]?\d+(?!e)'
        t.value = int(t.value)
        return t

    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += len(t.value)

    t_ignore = ' \t'

    def t_error(self, t):
        raise RuntimeError('invalid syntax: {}'.format(t))

    def __init__(self, filename='ifspec.ifs'):
        lex_dir = op.join(op.dirname(__file__), 'tables')
        lex_module = 'gomjabbar.ifs.tables.lextab'

        self.lexer = lex.lex(
            module=self, reflags=re.VERBOSE | re.IGNORECASE,
            lextab=lex_module, outputdir=lex_dir, optimize=1,
        )
        self.filename = filename
        self.lexer.filename = filename



from __future__ import print_function

import os.path as op

import ply.yacc as yacc

from .ast import Dash, Ifs, Range, Table, TableRow
from .lexer import IfsLexer


def _handle_list_reduction(p):
    items = p[1]
    if not isinstance(items, list):
        items = [items]
    p[0] = items


def _handle_list_or_empty_reduction(p):
    if len(p) > 2:
        items = p[2]
        if items is None:
            items = []
        elif not isinstance(items, list):
            items = [items]
        head = p[1]
        if head is None:
            head = []
        p[0] = head + items


class IfsParser(object):
    """ Parser for .ifs files
    """
    tokens = IfsLexer.tokens

    def p_ifs_table(self, p):
        """ ifs_file : list_of_tables """
        p[0] = Ifs(p[1])

    def p_list_of_tables1(self, p):
        """ list_of_tables : list_of_tables table """
        items = p[2]
        if not isinstance(items, list):
            items = [items]
        p[0] = p[1] + items

    def p_list_of_tables2(self, p):
        """ list_of_tables : table """
        _handle_list_reduction(p)

    def p_table(self, p):
        """ table : NAME_TABLE name_table
                  | PORT_TABLE port_table
                  | PARAMETER_TABLE parameter_table
                  | STATIC_VAR_TABLE static_var_table """
        p[0] = Table(p[1], p[2], lineno=p.lineno(1))

    def p_name_table(self, p):
        """ name_table : name_table name_table_item
                       | empty """
        _handle_list_or_empty_reduction(p)

    def p_name_table_item(self, p):
        """ name_table_item : C_FUNCTION_NAME identifier
                            | SPICE_MODEL_NAME identifier
                            | DESCRIPTION string """
        p[0] = TableRow(p[1], p[2], lineno=p.lineno(1))

    def p_port_table(self, p):
        """ port_table : port_table port_table_item
                       | empty """
        _handle_list_or_empty_reduction(p)

    def p_port_table_item(self, p):
        """ port_table_item : PORT_NAME list_of_ids
                            | DESCRIPTION list_of_strings
                            | DIRECTION list_of_directions
                            | DEFAULT_TYPE list_of_ctypes
                            | ALLOWED_TYPES list_of_ctype_lists
                            | ARRAY list_of_bool
                            | ARRAY_BOUNDS list_of_array_bounds
                            | NULL_ALLOWED list_of_bool """
        p[0] = TableRow(p[1], p[2], lineno=p.lineno(1))

    def p_parameter_table(self, p):
        """ parameter_table : parameter_table parameter_table_item
                            | empty """
        _handle_list_or_empty_reduction(p)

    def p_parameter_table_item(self, p):
        """ parameter_table_item : PARAMETER_NAME list_of_ids
                                 | DESCRIPTION list_of_strings
                                 | DATA_TYPE list_of_dtypes
                                 | DEFAULT_VALUE list_of_values
                                 | LIMITS list_of_ranges
                                 | ARRAY list_of_bool
                                 | ARRAY_BOUNDS list_of_array_bounds
                                 | NULL_ALLOWED list_of_bool """
        p[0] = TableRow(p[1], p[2], lineno=p.lineno(1))

    def p_static_var_table(self, p):
        """ static_var_table : static_var_table static_var_table_item
                             | empty """
        _handle_list_or_empty_reduction(p)

    def p_static_var_table_item(self, p):
        """ static_var_table_item : STATIC_VAR_NAME list_of_ids
                                  | DESCRIPTION list_of_strings
                                  | DATA_TYPE list_of_dtypes
                                  | ARRAY list_of_bool """
        p[0] = TableRow(p[1], p[2], lineno=p.lineno(1))

    def p_list_of_ids(self, p):
        """ list_of_ids : list_of_ids identifier
                        | empty """
        _handle_list_or_empty_reduction(p)

    def p_list_of_array_bounds(self, p):
        """ list_of_array_bounds : list_of_array_bounds int_range
                                 | list_of_array_bounds identifier
                                 | empty """
        _handle_list_or_empty_reduction(p)

    def p_list_of_strings(self, p):
        """ list_of_strings : list_of_strings string
                            | empty """
        _handle_list_or_empty_reduction(p)

    def p_list_of_directions(self, p):
        """ list_of_directions : list_of_directions direction
                               | empty """
        _handle_list_or_empty_reduction(p)

    def p_direction(self, p):
        """ direction : DIR_IN
                      | DIR_OUT
                      | DIR_INOUT """
        p[0] = p[1]

    def p_list_of_bool(self, p):
        """ list_of_bool : list_of_bool bool
                         | empty """
        _handle_list_or_empty_reduction(p)

    def p_list_of_ctypes(self, p):
        """ list_of_ctypes : list_of_ctypes ctype
                           | empty """
        _handle_list_or_empty_reduction(p)

    def p_ctype(self, p):
        """ ctype : CTYPE_V
                  | CTYPE_VD
                  | CTYPE_VNAM
                  | CTYPE_I
                  | CTYPE_ID
                  | CTYPE_G
                  | CTYPE_GD
                  | CTYPE_H
                  | CTYPE_HD
                  | CTYPE_D
                  | identifier """
        p[0] = p[1]

    def p_list_of_dtypes(self, p):
        """ list_of_dtypes : list_of_dtypes dtype
                           | empty """
        _handle_list_or_empty_reduction(p)

    def p_dtype(self, p):
        """ dtype : DTYPE_REAL
                  | DTYPE_INT
                  | DTYPE_BOOLEAN
                  | DTYPE_COMPLEX
                  | DTYPE_STRING
                  | DTYPE_POINTER """
        p[0] = p[1]

    def p_list_of_ranges(self, p):
        """ list_of_ranges : list_of_ranges range
                           | empty """
        _handle_list_or_empty_reduction(p)

    def p_int_range1(self, p):
        """ int_range : DASH """
        p[0] = Range(lineno=p.lineno(1))

    def p_int_range2(self, p):
        """ int_range : LBRACKET int_or_dash maybe_comma int_or_dash RBRACKET
        """
        p[0] = Range(p[2], p[4], lineno=p.lineno(1))

    def p_maybe_comma(self, p):
        """ maybe_comma : COMMA
                        | empty """
        pass

    def p_int_or_dash1(self, p):
        """ int_or_dash : DASH """
        p[0] = Dash(lineno=p.lineno(1))

    def p_int_or_dash2(self, p):
        """ int_or_dash : integer_value """
        p[0] = p[1]

    def p_range1(self, p):
        """ range : DASH """
        p[0] = Range(lineno=p.lineno(1))

    def p_range2(self, p):
        """ range : LBRACKET number_or_dash maybe_comma number_or_dash RBRACKET
        """
        p[0] = Range(p[2], p[4], lineno=p.lineno(1))

    def p_number_or_dash1(self, p):
        """ number_or_dash : DASH """
        p[0] = Dash(lineno=p.lineno(1))

    def p_number_or_dash2(self, p):
        """ number_or_dash : number """
        p[0] = p[1]

    def p_list_of_values(self, p):
        """ list_of_values : list_of_values value_or_dash
                           | empty """
        _handle_list_or_empty_reduction(p)

    def p_value_or_dash1(self, p):
        """ value_or_dash : DASH """
        p[0] = Dash(lineno=p.lineno(1))

    def p_value_or_dash2(self, p):
        """ value_or_dash : value """
        p[0] = p[1]

    def p_value(self, p):
        """ value : string
                  | bool
                  | complex
                  | number """
        p[0] = p[1]

    def p_complex(self, p):
        """ complex : LANGLE real maybe_comma real RANGLE """
        p[0] = complex(p[2], p[4])

    def p_list_of_ctype_lists(self, p):
        """ list_of_ctype_lists : list_of_ctype_lists delimited_ctype_list
                                | empty """
        _handle_list_or_empty_reduction(p)

    def p_delimited_ctype_list(self, p):
        """ delimited_ctype_list : LBRACKET ctype_list RBRACKET """
        p[0] = [p[2]]

    def p_ctype_list1(self, p):
        """ ctype_list : ctype """
        _handle_list_reduction(p)

    def p_ctype_list2(self, p):
        """ ctype_list : ctype_list maybe_comma ctype """
        items = p[3]
        if not isinstance(items, list):
            items = [items]
        p[0] = p[1] + items

    def p_bool(self, p):
        """ bool : BOOL_YES
                 | BOOL_NO """
        p[0] = p[1]

    def p_string(self, p):
        """ string : STRING_LITERAL """
        p[0] = p[1]

    def p_identifier(self, p):
        """ identifier : IDENTIFIER """
        p[0] = p[1]

    def p_number(self, p):
        """ number : real
                   | integer_value """
        p[0] = p[1]

    def p_integer_value(self, p):
        """ integer_value : integer """
        p[0] = p[1]

    def p_real(self, p):
        """ real : REAL_LITERAL """
        p[0] = p[1]

    def p_integer(self, p):
        """ integer : INT_LITERAL """
        p[0] = p[1]

    def p_empty(self, p):
        """ empty : """
        pass

    def p_error(self, t):
        msg = 'invalid syntax: {} ({})'
        raise RuntimeError(msg.format(t, t.lineno))

    def __init__(self):
        parse_dir = op.join(op.dirname(__file__), 'tables')
        parse_mod = 'gomjabbar.ifs.tables.parsetab'

        self.parser = yacc.yacc(
            module=self,
            method='LALR',
            start='ifs_file',
            tabmodule=parse_mod,
            outputdir=parse_dir,
            optimize=1,
            debug=0
        )

    def parse(self, source, filename='ifspec.ifs'):
        """ Parse source string and create abstract syntax tree (AST).
        """
        self.lexer = IfsLexer(filename)
        return self.parser.parse(source, debug=0, lexer=self.lexer.lexer)



import contextlib
import os
import os.path as op
from subprocess import check_call
import sys

import jinja2

from gomjabbar.ifs.build import (
    build_connections_list, build_parameters_list, count_static_vars,
    parse_file
)

BIN_DIR = op.join(op.abspath(sys.prefix), 'bin')
INCLUDE_DIR = op.join(op.abspath(sys.prefix), 'include')

DATA_DIR = op.join(op.dirname(__file__), 'data')
TEMPLATE_LOADER = jinja2.FileSystemLoader(DATA_DIR)
TEMPLATE_ENV = jinja2.Environment(loader=TEMPLATE_LOADER, trim_blocks=True,
                                  keep_trailing_newline=True)
TEMPLATE_ENV.filters['array_size'] = lambda v: max(1, len(v))


@contextlib.contextmanager
def build_test(dir_path, name, code, parameters_dict):
    """ Build a C source file which can be used to test part of a code model.
    """
    conns, params, num_vars = _get_template_context(dir_path, parameters_dict)
    output = op.join(dir_path, name + '.mod')

    try:
        with open(output, 'w') as fp:
            _render_templates(fp, conns, params, num_vars)
            fp.write(code)

        yield _compile_test(output)
    finally:
        _clean_test(op.join(dir_path, name))


def _clean_test(path):
    """ Clean up after a call to _compile_test
    """
    path = op.abspath(path)
    remove_paths = [path + ext for ext in ('', '.c', '.o', '.mod')]
    for path in remove_paths:
        if op.exists(path):
            os.remove(path)


def _compile_test(path):
    """ Build an executable for a code model test.

    Returns the path of the resulting executable.
    """
    def _compile_obj(p):
        base_name = op.splitext(p)[0]
        c_file = base_name + '.c'
        obj_file = base_name + '.o'
        cmd = ['cc', '-c', '-o', obj_file, c_file, '-I' + INCLUDE_DIR]
        check_call(cmd)
        return obj_file

    def _preprocess_mod(p):
        cmd = [op.join(BIN_DIR, 'cmpp'), '-mod', p]
        check_call(cmd)

    code_dir, path = op.split(path)
    module_path = op.splitext(path)[0]
    code_dir = op.abspath(code_dir)
    with _in_dir(code_dir):
        # Preprocess the .mod file with cmpp
        _preprocess_mod(path)
        # Compile the resulting .c file
        obj_file = _compile_obj(path)

        # Make sure cfunc.o exists
        cfunc_obj_file = 'cfunc.o'
        if not op.exists(cfunc_obj_file):
            _preprocess_mod('cfunc.mod')
            cfunc_obj_file = _compile_obj('cfunc.c')

        # Link the resulting .o file with cfunc.o into an executable
        cmd = ['cc', '-o', module_path, cfunc_obj_file, obj_file]
        check_call(cmd)

    return op.join(code_dir, module_path)


def _get_template_context(dir_path, parameters_dict):
    ast = parse_file(op.join(dir_path, 'ifspec.ifs'))
    connections, parameters = [], []
    num_static_vars = 0
    if ast.port_table:
        connections = build_connections_list(ast.port_table)
    if ast.parameter_table:
        parameters = build_parameters_list(ast.parameter_table,
                                           parameters_dict)
    if ast.static_var_table:
        num_static_vars = count_static_vars(ast.static_var_table)
    return connections, parameters, num_static_vars


@contextlib.contextmanager
def _in_dir(path):
    restore = op.abspath(op.curdir)
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(restore)


def _render_templates(fp, connections, parameters, num_static_vars):
    template = TEMPLATE_ENV.get_template('alloc.c.jinja')
    context = {
        'num_conns': len(connections),
        'num_params': len(parameters),
        'num_static_vars': num_static_vars,
    }
    fp.write(template.render(context))

    template = TEMPLATE_ENV.get_template('connections.c.jinja')
    context = {'connections': connections}
    fp.write(template.render(context))

    template = TEMPLATE_ENV.get_template('params.c.jinja')
    context = {'parameters': parameters}
    fp.write(template.render(context))

    macros_source = TEMPLATE_LOADER.get_source(TEMPLATE_ENV, 'macros.c')[0]
    fp.write(macros_source)



# Copyright (c) 2018 by Enthought, Inc.
# All rights reserved.
import os
import re
import subprocess

from setuptools import setup

MAJOR = 0
MINOR = 1
MICRO = 0

IS_RELEASED = True
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)


def git_version():
    """ Return the git revision as a string
    """
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, env=env,
        ).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'describe', '--tags'])
    except OSError:
        out = ''

    git_description = out.strip().decode('ascii')
    expr = r'.*?\-(?P<count>\d+)-g(?P<hash>[a-fA-F0-9]+)'
    match = re.match(expr, git_description)
    if match is None:
        git_revision, git_count = 'Unknown', '0'
    else:
        git_revision, git_count = match.group('hash'), match.group('count')

    return git_revision, git_count


def write_version_py(filename='gomjabbar/_version.py'):
    template = """\
# THIS FILE IS GENERATED FROM GOMJABBAR SETUP.PY
version = '{version}'
full_version = '{full_version}'
git_revision = '{git_revision}'
is_released = {is_released}

if not is_released:
    version = full_version
"""
    # Adding the git rev number needs to be done inside
    # write_version_py(), otherwise the import of gomjabbar._version messes
    # up the build under Python 3.
    fullversion = VERSION
    if os.path.exists('.git'):
        git_rev, dev_num = git_version()
    elif os.path.exists('gomjabbar/_version.py'):
        # must be a source distribution, use existing version file
        try:
            from gomjabbar._version import git_revision as git_rev
            from gomjabbar._version import full_version as full_v
        except ImportError:
            raise ImportError("Unable to import git_revision. Try removing "
                              "gomjabbar/_version.py and the build directory "
                              "before building.")

        match = re.match(r'.*?\.dev(?P<dev_num>\d+)', full_v)
        if match is None:
            dev_num = '0'
        else:
            dev_num = match.group('dev_num')
    else:
        git_rev = 'Unknown'
        dev_num = '0'

    if not IS_RELEASED:
        fullversion += '.dev{0}'.format(dev_num)

    with open(filename, "wt") as fp:
        fp.write(template.format(version=VERSION,
                                 full_version=fullversion,
                                 git_revision=git_rev,
                                 is_released=IS_RELEASED))


if __name__ == "__main__":
    write_version_py()
    setup()
